#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/fs.h>
#include <signal.h> 
#include <unistd.h>
#include <wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h> 
#include <sys/shm.h> 
#include <sys/time.h>

int diskfd = -1;
int shm    = -1; 	/* shared memory ID*/
long long readTotal = 0;

void terminate()
{
	if (diskfd >= 0) {
		close(diskfd);
		diskfd = -1;
	}
}

int main(int argc, char* argv[])
{
	unsigned long read_size = 1073741824;
	int block_size = 16384;
	int processNum = 3;

	/* allocate shared memory */
	int shm = shmget(IPC_PRIVATE, 8, IPC_CREAT);
	if (shm < 0) {
		printf("Faild shmget %d\n", errno );
		return -1;
	}
	
	/* file open */
	if ((diskfd = open(argv[1], O_RDONLY )) < 0) {
		printf("Cannot open %s\n", argv[1]);
		return -1;
	}

	struct timeval curtimeval         = {0};
	struct timeval restimeval         = {0};
	struct timeval starttimeval       = {0};
	struct timeval each_process_res   = {0};
	struct timeval each_process_start = {0};
	struct timeval each_process_end   = {0};

	int myProcessIndex = 0;
	int processCount = 0;

	for (;processCount < processNum ;processCount++) {
		int pid = fork();
		if (pid == 0) { /* childern  */
			myProcessIndex = processCount;
			break;
		} else if (pid < 0){
			printf("fork error (errno=%d)\n",errno);
			terminate();
			exit(-1);
		}
	}

	/* get start time */
	gettimeofday(&starttimeval,NULL);

	if (processCount == processNum) {

	/*
	 *   Here is processing of parent. 
	 */
		int status;
		int processCount = 0;
	
		for(;processCount < processNum ;processCount++)
			wait(&status);
		
		long *pReadTotal = shmat(shm,0,0);
	
		/* Calculate　average  */
		gettimeofday(&curtimeval,NULL);
		timersub(&curtimeval, &starttimeval, &restimeval);
		float totaltime = (float)restimeval.tv_sec
				  + ((float)restimeval.tv_usec / 1000000.);
		printf("***Total Ave %f MB/sec ***\n",
		       ((*pReadTotal)/(1024*1024))/totaltime);
		
		shmdt(pReadTotal);
		shmctl(shm,IPC_RMID,0);

		return 0;
	}

/*
 *   The processing of children since this. 
 */
	char *buf = valloc(block_size);
	unsigned int count = 0;
	float each_process_time;

	gettimeofday(&each_process_start,NULL);
	for (;;count++) {

		/* Each process reads a separate sector. */
		unsigned long long seekpos = (processNum*block_size*count) + (myProcessIndex*block_size);
		unsigned int needsize = block_size;

		if (seekpos >= read_size)
			break;

		if (lseek(diskfd, seekpos, 0) == -1) {
			printf("Seek Faild!!.%d\n",errno);
			break;
		}
		
		/* read */
		int readSize = read(diskfd, buf, needsize);
		if (readSize != needsize) {
			break;
		}
		readTotal += readSize;
	}
	gettimeofday(&each_process_end,NULL);
	timersub(&each_process_end, &each_process_start, &each_process_res);
	each_process_time = (float)each_process_res.tv_sec
			  + ((float)each_process_res.tv_usec / 1000000.);

	free(buf);
	
	long long *pReadTotal = shmat(shm,0,0);
	*pReadTotal += readTotal;
	shmdt(pReadTotal);

	return 0;
}
